/**
 * Copyright (C) 2003 Michel Santos
 * Licensed under the Academic Free License version 3.0
 * Please see the LICENSE for more information.
 *
 * JUnit based test
 *
 * Created on May 26, 2003, 2:06 PM
 */

package cern.colt.matrix.linalg;


import junit.framework.*;
import cern.colt.matrix.*;


/**
 *
 * @author Michel Santos
 */
public class CheckerTest extends TestCase {
  
  public CheckerTest(String testName) {
    super(testName);
  }

  protected void setUp() throws Exception {
  }

  protected void tearDown() throws Exception {
  }

  public static Test suite() {
    TestSuite suite = new TestSuite(CheckerTest.class);
    
    return suite;
  }

  private cern.colt.matrix.linalg.Algebra2 LA = new cern.colt.matrix.linalg.Algebra2();
  private DoubleFactory2D factory2D = DoubleFactory2D2.dense;
  private DoubleFactory1D factory1D = DoubleFactory1D.dense;  

  protected void print(int[] array) {
    for (int i = 0; i < array.length; i++) {
      System.out.println("Index #" + i + " = " + array[i]);
    }
  }  
  
  
  public void testConsistency1() {
    System.out.println();
    System.out.println("**********");
    System.out.println("* Consistency Check #1 - Inconsistent");
    System.out.println("**********");
    System.out.println();
    
    double[][] dA = new double[][] { 
        {1875, 25, 25, 25}
      , {25, 1, 0, 0}
      , {25, 0, 1, 0}
      , {25, 0, 0, 1}
    };
    double[] dB = new double[] {-31470.125, -385.275, -452.7845, -362.7455};
    
    DoubleMatrix2D matA = factory2D.make(dA);
    DoubleMatrix1D vecB = factory1D.make(dB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    Checker.scaleEquality(matA, vecB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    assertTrue(!Checker.isConsistent(matA, vecB));
  }
  
  public void testConsistency2() {
    System.out.println();
    System.out.println("**********");
    System.out.println("* Consistency Check #2 - Consistent");
    System.out.println("**********");
    System.out.println();
    
    // x = 1
    // y = 2
    // z = 3
    double[][] dA = new double[][] { 
        {1, 0, 0}
      , {0, 1, 0}
      , {0, 0, 1}
    };
    double[] dB = new double[] {1, 2, 3};
    
    DoubleMatrix2D matA = factory2D.make(dA);
    DoubleMatrix1D vecB = factory1D.make(dB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    Checker.scaleEquality(matA, vecB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    assertTrue(Checker.isConsistent(matA, vecB));
  }
  
  public void testConsistency3() {
    System.out.println();
    System.out.println("**********");
    System.out.println("* Consistency Check #3 - Consistent but redundant");
    System.out.println("**********");
    System.out.println();
    
    // x = 1
    // y = 2
    // z = 3
    // 2*y = 2*2
    double[][] dA = new double[][] { 
        {1, 0, 0}
      , {0, 1, 0}
      , {0, 0, 1}
      , {0, 2*1, 0}
    };
    double[] dB = new double[] {1, 2, 3, 2*2};
    
    DoubleMatrix2D matA = factory2D.make(dA);
    DoubleMatrix1D vecB = factory1D.make(dB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    Checker.scaleEquality(matA, vecB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    assertTrue(Checker.isConsistent(matA, vecB));
  }

  public void testConsistency4() {
    System.out.println();
    System.out.println("**********");
    System.out.println("* Consistency Check #4 - Consistent but too few equations");
    System.out.println("**********");
    System.out.println();
    
    // 2x = 33
    // 3y = 44
    double[][] dA = new double[][] { 
        {2, 0, 0}
      , {0, 3, 0}
    };
    double[] dB = new double[] {33, 44};
    
    DoubleMatrix2D matA = factory2D.make(dA);
    DoubleMatrix1D vecB = factory1D.make(dB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    Checker.scaleEquality(matA, vecB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    assertTrue(Checker.isConsistent(matA, vecB));
  }  

  public void testConsistency5() {
    System.out.println();
    System.out.println("**********");
    System.out.println("* Consistency Check #5 - Consistent but too few equations");
    System.out.println("**********");
    System.out.println();
    
    // -3.14z = 52
    double[][] dA = new double[][] { 
      {0, 0, -3.14}
    };
    double[] dB = new double[] {52};
    
    DoubleMatrix2D matA = factory2D.make(dA);
    DoubleMatrix1D vecB = factory1D.make(dB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    Checker.scaleEquality(matA, vecB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    assertTrue(Checker.isConsistent(matA, vecB));
  }    

  public void testConsistency6() {
    System.out.println();
    System.out.println("**********");
    System.out.println("* Consistency Check #6 - Inconsistent and redundant");
    System.out.println("**********");
    System.out.println();
    
    // x = 5
    // x = 5
    // x = 519
    double[][] dA = new double[][] { 
      {1, 0, 0}
      , {1, 0, 0}
      , {1, 0, 0}
    };
    double[] dB = new double[] {5, 5, 519};
    
    DoubleMatrix2D matA = factory2D.make(dA);
    DoubleMatrix1D vecB = factory1D.make(dB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    Checker.scaleEquality(matA, vecB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    assertTrue(!Checker.isConsistent(matA, vecB));
  }    
  
  public void testConsistency7() {
    System.out.println();
    System.out.println("**********");
    System.out.println("* Consistency Check #7 - Hardly inconsistent");
    System.out.println("**********");
    System.out.println();
    
    // x = -5
    // y = -7
    // z = -13
    // x = -5 + 1e-15
    double[][] dA = new double[][] { 
      {1, 0, 0}
      , {0, 1, 0}
      , {0, 0, 1}
      , {1, 0, 0}
    };
    double[] dB = new double[] {-5, -7, -13, -5 + 1e-15};
    
    DoubleMatrix2D matA = factory2D.make(dA);
    DoubleMatrix1D vecB = factory1D.make(dB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    Checker.scaleEquality(matA, vecB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    assertTrue(Checker.isConsistent(matA, vecB));
  }

  
  public void testConsistency8() {
    System.out.println();
    System.out.println("**********");
    System.out.println("* Consistency Check #8 - Slightly inconsistent");
    System.out.println("**********");
    System.out.println();
    
    // x = -5
    // y = -7
    // z = -13
    // x = -5 + 1e-10
    double[][] dA = new double[][] { 
      {1, 0, 0}
      , {0, 1, 0}
      , {0, 0, 1}
      , {1, 0, 0}
    };
    double[] dB = new double[] {-5, -7, -13, -5 + 1e-10};
    
    DoubleMatrix2D matA = factory2D.make(dA);
    DoubleMatrix1D vecB = factory1D.make(dB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    Checker.scaleEquality(matA, vecB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    assertTrue(!Checker.isConsistent(matA, vecB));
  }

  public void testConsistency9() {
    System.out.println();
    System.out.println("**********");
    System.out.println("* Consistency Check #9 - Consistent");
    System.out.println("**********");
    System.out.println();
    
    // x + y + z = 1
    // x + 2y + 3z = -55
    // -5x -44y + 14z = 2
    double[][] dA = new double[][] { 
        {1, 1, 1}
      , {1, 2, 3}
      , {-5, -44, 14}
    };
    double[] dB = new double[] {1, -55, 2};
    
    DoubleMatrix2D matA = factory2D.make(dA);
    DoubleMatrix1D vecB = factory1D.make(dB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    Checker.scaleEquality(matA, vecB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    assertTrue(Checker.isConsistent(matA, vecB));
  }
  
  public void testConsistency10() {
    System.out.println();
    System.out.println("**********");
    System.out.println("* Consistency Check #10 - Inconsistent");
    System.out.println("**********");
    System.out.println();
    
    // x + y + z = 1
    // z = 99
    // x + 2y + 3z = -55
    // -5x -44y + 14z = 2    
    double[][] dA = new double[][] { 
        {1, 1, 1}
      , {0, 0, 1}
      , {1, 2, 3}
      , {-5, -44, 14}
    };
    double[] dB = new double[] {1, 99, -55, 2};
    
    DoubleMatrix2D matA = factory2D.make(dA);
    DoubleMatrix1D vecB = factory1D.make(dB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    Checker.scaleEquality(matA, vecB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    assertTrue(!Checker.isConsistent(matA, vecB));
  }
  
  public void testConsistency11() {
    System.out.println();
    System.out.println("**********");
    System.out.println("* Consistency Check #11 - Consistent but redundant");
    System.out.println("**********");
    System.out.println();
    
    // x + y + z = 1
    // -33x -33y -33z = -33
    // x + 2y + 3z = -55
    // -5x -44y + 14z = 2    
    double[][] dA = new double[][] { 
        {1, 1, 1}
      , {-33, -33, -33}
      , {1, 2, 3}
      , {-5, -44, 14}
    };
    double[] dB = new double[] {1, -33, -55, 2};
    
    DoubleMatrix2D matA = factory2D.make(dA);
    DoubleMatrix1D vecB = factory1D.make(dB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    Checker.scaleEquality(matA, vecB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    assertTrue(Checker.isConsistent(matA, vecB));
  }  

  public void testConsistency12() {
    System.out.println();
    System.out.println("**********");
    System.out.println("* Consistency Check #12 - Consistent but very redundant");
    System.out.println("**********");
    System.out.println();
    
    // x + y + z = 1
    // -33x -33y -33z = -33
    // x + 2y + 3z = -55
    // -5x -44y + 14z = 2
    // x + 2y + 3z = -55
    // -5x -44y + 14z = 2
    // x + y + z = 1
    double[][] dA = new double[][] { 
        {1, 1, 1}
      , {-33, -33, -33}
      , {1, 2, 3}
      , {-5, -44, 14}
      , {1, 2, 3}
      , {-5, -44, 14}
      , {1, 1, 1}
    };
    double[] dB = new double[] {1, -33, -55, 2, -55, 2, 1};
    
    DoubleMatrix2D matA = factory2D.make(dA);
    DoubleMatrix1D vecB = factory1D.make(dB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    Checker.scaleEquality(matA, vecB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    assertTrue(Checker.isConsistent(matA, vecB));
  }  
  

  public void testConsistency13() {
    System.out.println();
    System.out.println("**********");
    System.out.println("* Consistency Check #13 - Inconsistent and very redundant");
    System.out.println("**********");
    System.out.println();
    
    // x + y + z = 1
    // -33x -33y -33z = -33
    // x + 2y + 3z = -55
    // -5x -44y + 14z = 2
    // x + y + z = 123 <-- This is inconsistent
    // x + 2y + 3z = -55
    // -5x -44y + 14z = 2
    // x + y + z = 1    
    double[][] dA = new double[][] { 
        {1, 1, 1}
      , {-33, -33, -33}
      , {1, 2, 3}
      , {-5, -44, 14}
      , {1, 1, 1}
      , {1, 2, 3}
      , {-5, -44, 14}
      , {1, 1, 1}
    };
    double[] dB = new double[] {1, -33, -55, 2, 123, -55, 2, 1};
    
    DoubleMatrix2D matA = factory2D.make(dA);
    DoubleMatrix1D vecB = factory1D.make(dB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    Checker.scaleEquality(matA, vecB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    assertTrue(!Checker.isConsistent(matA, vecB));
  }  

  
  
  public void testRedundancy1() {
    System.out.println();
    System.out.println("**********");
    System.out.println("* Redundancy Check #1 - Consistent");
    System.out.println("**********");
    System.out.println();
    
    // x = 1
    // y = 2
    // z = 3
    double[][] dA = new double[][] { 
        {1, 0, 0}
      , {0, 1, 0}
      , {0, 0, 1}
    };
    double[] dB = new double[] {1, 2, 3};
    
    DoubleMatrix2D matA = factory2D.make(dA);
    DoubleMatrix1D vecB = factory1D.make(dB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    Checker.scaleEquality(matA, vecB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    assertTrue(Checker.isConsistent(matA, vecB));

    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    assertTrue(!Checker.isRedundant(matA));
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);    
  }
  
  public void testRedundancy2() {
    System.out.println();
    System.out.println("**********");
    System.out.println("* Redundancy Check #2 - Consistent but redundant");
    System.out.println("**********");
    System.out.println();
    
    // x = 1
    // 44x = 44
    // y = 2
    // z = 3    
    double[][] dA = new double[][] { 
        {1, 0, 0}
      , {44, 0, 0}
      , {0, 1, 0}
      , {0, 0, 1}
    };
    double[] dB = new double[] {1, 44, 2, 3};
    
    DoubleMatrix2D matA = factory2D.make(dA);
    DoubleMatrix1D vecB = factory1D.make(dB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    Checker.scaleEquality(matA, vecB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    assertTrue(Checker.isConsistent(matA, vecB));

    assertTrue(Checker.isRedundant(matA));
  }
  
  public void testRedundancy3() {
    System.out.println();
    System.out.println("**********");
    System.out.println("* Redundancy Check #3 - Inconsistent but redundant");
    System.out.println("**********");
    System.out.println();
    
    // x = 1
    // x = 0
    // y = 2
    // z = 3    
    double[][] dA = new double[][] { 
        {1, 0, 0}
      , {1, 0, 0}
      , {0, 1, 0}
      , {0, 0, 1}
    };
    double[] dB = new double[] {1, 0, 2, 3};
    
    DoubleMatrix2D matA = factory2D.make(dA);
    DoubleMatrix1D vecB = factory1D.make(dB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    Checker.scaleEquality(matA, vecB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);

    assertTrue(!Checker.isConsistent(matA, vecB));

    // assertTrue(isRedundant(matA, vecB));
  }  
  
  public void testRedundancy4() {
    System.out.println();
    System.out.println("**********");
    System.out.println("* Redundancy Check #4 - Consistent but redundant");
    System.out.println("**********");
    System.out.println();
    
    // x = 1
    // 44x = 44
    // y = 2
    // 34y = 68
    // z = 3    
    double[][] dA = new double[][] { 
        {1, 0, 0}
      , {44, 0, 0}
      , {0, 1, 0}
      , {0, 34, 0}
      , {0, 0, 1}
    };
    double[] dB = new double[] {1, 44, 2, 68, 3};
    
    DoubleMatrix2D matA = factory2D.make(dA);
    DoubleMatrix1D vecB = factory1D.make(dB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    Checker.scaleEquality(matA, vecB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);

    assertTrue(Checker.isConsistent(matA, vecB));

    assertTrue(Checker.isRedundant(matA));
  }

  
  public void testRedundancy5() {
    System.out.println();
    System.out.println("**********");
    System.out.println("* Redundancy Check #5 - Consistent but very redundant");
    System.out.println("**********");
    System.out.println();
    
    // x + y + z = 1
    // -33x -33y -33z = -33
    // x + 2y + 3z = -55
    // -5x -44y + 14z = 2
    // x + 2y + 3z = -55
    // -5x -44y + 14z = 2
    // x + y + z = 1
    double[][] dA = new double[][] { 
        {1, 1, 1}
      , {-33, -33, -33}
      , {1, 2, 3}
      , {-5, -44, 14}
      , {1, 2, 3}
      , {-5, -44, 14}
      , {1, 1, 1}
    };
    double[] dB = new double[] {1, -33, -55, 2, -55, 2, 1};
    
    DoubleMatrix2D matA = factory2D.make(dA);
    DoubleMatrix1D vecB = factory1D.make(dB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    Checker.scaleEquality(matA, vecB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    assertTrue(Checker.isConsistent(matA, vecB));
    
    assertTrue(Checker.isRedundant(matA));
  }  

  
  public void testRedundancy6() {
    System.out.println();
    System.out.println("**********");
    System.out.println("* Redundancy Check #6 - Consistent but too few equations");
    System.out.println("**********");
    System.out.println();
    
    // 2x = 33
    // 3y = 44
    double[][] dA = new double[][] { 
        {2, 0, 0}
      , {0, 3, 0}
    };
    double[] dB = new double[] {33, 44};
    
    DoubleMatrix2D matA = factory2D.make(dA);
    DoubleMatrix1D vecB = factory1D.make(dB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    Checker.scaleEquality(matA, vecB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    assertTrue(Checker.isConsistent(matA, vecB));
    
    assertTrue(!Checker.isRedundant(matA));
  }

  public void testRedundancy7() {
    System.out.println();
    System.out.println("**********");
    System.out.println("* Redundancy Check #7 - Consistent but too few equations");
    System.out.println("**********");
    System.out.println();
    
    // 505z = -3192.45
    double[][] dA = new double[][] { 
        {0, 0, 505}
    };
    double[] dB = new double[] {-3192.45};
    
    DoubleMatrix2D matA = factory2D.make(dA);
    DoubleMatrix1D vecB = factory1D.make(dB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    Checker.scaleEquality(matA, vecB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    assertTrue(Checker.isConsistent(matA, vecB));
    
    assertTrue(!Checker.isRedundant(matA));
  }

  
  public void testRedundancy1_v2() throws InconsistentEquationsException {
    System.out.println();
    System.out.println("**********");
    System.out.println("* Redundancy Check #1 - Consistent");
    System.out.println("**********");
    System.out.println();
    
    // x = 1
    // y = 2
    // z = 3
    double[][] dA = new double[][] { 
        {1, 0, 0}
      , {0, 1, 0}
      , {0, 0, 1}
    };
    double[] dB = new double[] {1, 2, 3};
    
    DoubleMatrix2D matA = factory2D.make(dA);
    DoubleMatrix1D vecB = factory1D.make(dB);
    
    Checker checker = new Checker(matA, vecB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    assertTrue(checker.isConsistent());
    assertTrue(!checker.isRedundant());
    
    int[] iAryRedundantActual = checker.listRedundantEquations();
    int[] iAryRedundantExpected = new int[] {};
    checkRedundancies(iAryRedundantExpected, iAryRedundantActual);    
  }

  public void testRedundancy2_v2() throws InconsistentEquationsException {
    System.out.println();
    System.out.println("**********");
    System.out.println("* Redundancy Check #2 - Consistent but redundant");
    System.out.println("**********");
    System.out.println();
    
    // x = 1
    // 44x = 44
    // y = 2
    // z = 3    
    double[][] dA = new double[][] { 
        {1, 0, 0}
      , {44, 0, 0}
      , {0, 1, 0}
      , {0, 0, 1}
    };
    double[] dB = new double[] {1, 44, 2, 3};
    
    DoubleMatrix2D matA = factory2D.make(dA);
    DoubleMatrix1D vecB = factory1D.make(dB);
    
    Checker checker = new Checker(matA, vecB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    assertTrue(checker.isConsistent());
    assertTrue(checker.isRedundant());
    
    int[] iAryRedundantActual = checker.listRedundantEquations();
    int[] iAryRedundantExpected = new int[] {1};
    checkRedundancies(iAryRedundantExpected, iAryRedundantActual);    
  }

  
  public void testRedundancy3_v2() throws InconsistentEquationsException {
    System.out.println();
    System.out.println("**********");
    System.out.println("* Redundancy Check #3 - Inconsistent but redundant");
    System.out.println("**********");
    System.out.println();
    
    // x = 1
    // x = 0
    // y = 2
    // z = 3    
    double[][] dA = new double[][] { 
        {1, 0, 0}
      , {1, 0, 0}
      , {0, 1, 0}
      , {0, 0, 1}
    };
    double[] dB = new double[] {1, 0, 2, 3};
    
    DoubleMatrix2D matA = factory2D.make(dA);
    DoubleMatrix1D vecB = factory1D.make(dB);
    
    Checker checker = new Checker(matA, vecB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    assertTrue(!checker.isConsistent());
  }

  public void testRedundancy4_v2() throws InconsistentEquationsException {
    System.out.println();
    System.out.println("**********");
    System.out.println("* Redundancy Check #4 - Consistent but redundant");
    System.out.println("**********");
    System.out.println();
    
    // x = 1
    // 44x = 44
    // y = 2
    // 34y = 68
    // z = 3    
    double[][] dA = new double[][] { 
        {1, 0, 0}
      , {44, 0, 0}
      , {0, 1, 0}
      , {0, 34, 0}
      , {0, 0, 1}
    };
    double[] dB = new double[] {1, 44, 2, 68, 3};
    
    DoubleMatrix2D matA = factory2D.make(dA);
    DoubleMatrix1D vecB = factory1D.make(dB);
    
    Checker checker = new Checker(matA, vecB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    assertTrue(checker.isConsistent());
    assertTrue(checker.isRedundant());    
    
    int[] iAryRedundantActual = checker.listRedundantEquations();
    int[] iAryRedundantExpected = new int[] {1, 3};
    checkRedundancies(iAryRedundantExpected, iAryRedundantActual);
  }  
  
  public void testRedundancy5_v2() throws InconsistentEquationsException {
    System.out.println();
    System.out.println("**********");
    System.out.println("* Redundancy Check #5 - Consistent but very redundant");
    System.out.println("**********");
    System.out.println();
    
    // x + y + z = 1
    // -33x -33y -33z = -33
    // x + 2y + 3z = -55
    // -5x -44y + 14z = 2
    // x + 2y + 3z = -55
    // -5x -44y + 14z = 2
    // x + y + z = 1
    double[][] dA = new double[][] { 
        {1, 1, 1}
      , {-33, -33, -33}
      , {1, 2, 3}
      , {-5, -44, 14}
      , {1, 2, 3}
      , {-5, -44, 14}
      , {1, 1, 1}
    };
    double[] dB = new double[] {1, -33, -55, 2, -55, 2, 1};

    DoubleMatrix2D matA = factory2D.make(dA);
    DoubleMatrix1D vecB = factory1D.make(dB);
    
    Checker checker = new Checker(matA, vecB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    assertTrue(checker.isConsistent());
    assertTrue(checker.isRedundant());
    
    int[] iAryRedundantActual = checker.listRedundantEquations();
    int[] iAryRedundantExpected = new int[] {1, 4, 5, 6};
    checkRedundancies(iAryRedundantExpected, iAryRedundantActual);    
  }    
  
  public void testRedundancy6_v2() throws InconsistentEquationsException {
    System.out.println();
    System.out.println("**********");
    System.out.println("* Redundancy Check #6 - Consistent but too few equations");
    System.out.println("**********");
    System.out.println();
    
    // 2x = 33
    // 3y = 44
    double[][] dA = new double[][] { 
        {2, 0, 0}
      , {0, 3, 0}
    };
    double[] dB = new double[] {33, 44};
    
    DoubleMatrix2D matA = factory2D.make(dA);
    DoubleMatrix1D vecB = factory1D.make(dB);
    
    Checker checker = new Checker(matA, vecB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    assertTrue(checker.isConsistent());
    assertTrue(!checker.isRedundant());
    
    int[] iAryRedundantActual = checker.listRedundantEquations();
    int[] iAryRedundantExpected = new int[] {};
    checkRedundancies(iAryRedundantExpected, iAryRedundantActual);    
  }
  
  public void testRedundancy7_v2() throws InconsistentEquationsException {
    System.out.println();
    System.out.println("**********");
    System.out.println("* Redundancy Check #7 - Consistent but too few equations");
    System.out.println("**********");
    System.out.println();
    
    // 505z = -3192.45
    double[][] dA = new double[][] { 
        {0, 0, 505}
    };
    double[] dB = new double[] {-3192.45};
    
    DoubleMatrix2D matA = factory2D.make(dA);
    DoubleMatrix1D vecB = factory1D.make(dB);
    
    Checker checker = new Checker(matA, vecB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    assertTrue(checker.isConsistent());
    assertTrue(!checker.isRedundant());
    
    int[] iAryRedundantActual = checker.listRedundantEquations();
    int[] iAryRedundantExpected = new int[] {};
    checkRedundancies(iAryRedundantExpected, iAryRedundantActual);
  }
  
  private void checkRedundancies(int[] aryExpected, int[] aryActual) {
    int iSize = aryActual.length;
    if (iSize > 0) {
      System.out.println("Actual Redundant Equations: ");
      print(aryActual);
    }
    
    assertTrue(isRedundancyOK(aryExpected, aryActual));
  }
  
  private boolean isRedundancyOK(int[] aryExpected, int[] aryActual) {
    // Safety-check
    if (aryExpected.length != aryActual.length) {return false;}
    
    java.util.Arrays.sort(aryExpected);
    java.util.Arrays.sort(aryActual);
    
    // Loop through expected list
    int iSize = aryExpected.length;
    for (int i = 0; i < iSize; i++) {
      boolean bItemFound = java.util.Arrays.binarySearch(aryActual, aryExpected[i]) >= 0;
      if (!bItemFound) {return false;}
    }
    
    return true;
  }
  
  public void testLike() {
    double[] dB = new double[] {-3192.45, 55, -3};
    DoubleMatrix1D vecB = factory1D.make(dB);
    System.out.println("vecB: " + vecB);
    DoubleMatrix2D matB = vecB.like2D(3, 1);
    System.out.println("matB: " + matB);
    matB = factory2D.make(dB, 3);
    System.out.println("matB: " + matB);
  }
  
  
  public void testRankDeficientSolve1() throws InconsistentEquationsException {
    System.out.println();
    System.out.println("**********");
    System.out.println("* Rank Deficient Solver #1");
    System.out.println("**********");
    System.out.println();
    
    // x + 2y + 3z = 10
    // 4x + 5y + 6z = 9
    double[][] dA = new double[][] { 
        {1, 2, 3}
      , {4, 5, 6}
    };
    double[] dB = new double[] {10, 9};
    
    DoubleMatrix2D matA = factory2D.make(dA);
    DoubleMatrix1D vecB = factory1D.make(dB);
    
    Checker checker = new Checker(matA, vecB);
    
    System.out.println("matA: " + matA);
    System.out.println("vecB: " + vecB);
    
    assertTrue(checker.isConsistent());
    assertTrue(!checker.isRedundant());
    
    DoubleMatrix1D vecSolution = QRDecomposition2.solveRankDeficient(matA, vecB);
    System.out.println("Solution: " + vecSolution);
    DoubleMatrix1D vecError = LA.minus(LA.mult(matA, vecSolution), vecB);
    System.out.println("Error: " + vecError);
    
    double dTOL = 1e-12;
    assertEquals(-5.5, vecSolution.get(0), dTOL);
    assertEquals(0, vecSolution.get(1), dTOL);
    assertEquals(5.16666666666667, vecSolution.get(2), dTOL);
  }
  
  
}
