/**
 * Copyright (C) 2003 Michel Santos
 * Licensed under the Academic Free License version 3.0
 * Please see the LICENSE for more information.
 *
 * JUnit based test
 *
 * Created on May 26, 2003, 2:06 PM
 */

package cern.colt.matrix.linalg;


import cern.colt.matrix.*;


/**
 *
 * @author Michel Santos
 */
public class QRDecomposition2Test extends junit.framework.TestCase {
  
  public static void main(java.lang.String[] args) {
    junit.textui.TestRunner.run(suite());
  }
  
  public QRDecomposition2Test(java.lang.String testName) {
    super(testName);
  }
  
  public static junit.framework.Test suite() {
    junit.framework.TestSuite suite = new junit.framework.TestSuite(QRDecomposition2Test.class);
    return suite;
  }
  
  public void testOK() {
    System.out.println();
    System.out.println("*****");
    System.out.println("Simple test");
    System.out.println("*****");
    
    double[][] A = new double[][] {{0,3}, {2,0}, {0,0}};
    DoubleMatrix2D matA = factory2D.dense.make(A);
    System.out.println("A: " + matA);
    QRDecomposition2 qr = new QRDecomposition2(matA);
    System.out.println(qr);
  }
  
  
  private cern.colt.matrix.linalg.Algebra2 LA = new cern.colt.matrix.linalg.Algebra2();
  private DoubleFactory2D factory2D = DoubleFactory2D2.dense;
  private DoubleFactory1D factory1D = DoubleFactory1D.dense;  
  
  public void testHouseholderReflection_Ex_5_1_2() {
    System.out.println();
    System.out.println("*****");
    System.out.println("Householder test");
    System.out.println("*****");
    
    double[] x = new double[] {3, 1, 5, 1};
    DoubleMatrix1D vecX = DoubleFactory1D.dense.make(x);
    System.out.println("x: " + vecX);
    
    cern.colt.matrix.linalg.QRDecomposition2.HouseholderVector house = new cern.colt.matrix.linalg.QRDecomposition2.HouseholderVector(vecX);
    DoubleMatrix1D vecV = house.getReflection();
    System.out.println("Householder Reflector: " + vecV);
    System.out.println("Householder Multiplier: " + house.getMultiplier());
    
    DoubleMatrix2D matP = house.getReflector();
    System.out.println("Householder Reflection Matrix: " + matP);
    DoubleMatrix1D vecReflection = LA.mult(matP, vecX);
    System.out.println("Householder result: " + vecReflection);
    
    double dTOL = 1e-9;
    assertEquals(6, Math.abs(vecReflection.getQuick(0)), dTOL);
    assertEquals(0, vecReflection.getQuick(1), dTOL);
    assertEquals(0, vecReflection.getQuick(2), dTOL);
    assertEquals(0, vecReflection.getQuick(3), dTOL);
  }
  
  
  public void testAppendColumns() {
    System.out.println();
    System.out.println("*****");
    System.out.println("DoubleFactory2D Append Columns");
    System.out.println("*****");
    
    double[][] A = new double[][] { {3, 2, 1}, {6, 5, 1}, {9, 8, 1}, {12, 11, 1}};
    DoubleMatrix2D matA = factory2D.dense.make(A);
    System.out.println("A: " + matA);
    
    System.out.println();
    double[] aryX = new double[] {11, 22, 33, 44};
    DoubleMatrix1D vecX = DoubleFactory1D.dense.make(aryX);
    System.out.println("Vector: " + vecX);
    DoubleMatrix2D matCompound = DoubleFactory2D2.dense.appendColumns(matA, vecX);
    System.out.println("Compound matrix: " + matCompound);
    
    System.out.println();
    aryX = new double[] {11, 22, 33};
    vecX = DoubleFactory1D.dense.make(aryX);
    System.out.println("Vector: " + vecX);
    matCompound = DoubleFactory2D2.dense.appendColumns(matA, vecX);
    System.out.println("Compound matrix: " + matCompound);
    
    System.out.println();
    aryX = new double[] {11, 22, 33, 44, 55};
    vecX = DoubleFactory1D.dense.make(aryX);
    System.out.println("Vector: " + vecX);
    matCompound = DoubleFactory2D2.dense.appendColumns(matA, vecX);
    System.out.println("Compound matrix: " + matCompound);
    
  }
  
  public void testHouseholderQR_Ex_5_4_1_Test1() {
    System.out.println();
    System.out.println("*****");
    System.out.println("Householder QR Test #1");
    System.out.println("*****");
    
    double[][] A = new double[][] { {3, 2, 1}, {6, 5, 1}, {9, 8, 1}, {12, 11, 1}};
    DoubleMatrix2D matA = factory2D.dense.make(A);
    System.out.println("A: " + matA);
    
    QRDecomposition2 qr = new QRDecomposition2(matA);
    System.out.println(qr);
    
    DoubleMatrix2D matAComputed = LA.mult(qr.getQ(), qr.getR());
    System.out.println("A (computed): " + matAComputed);
    assertEquals(matA, matAComputed);
    checkDiagonalElements(qr.getR());
    
    printColumnIndependence(matA, qr);
  }
  
  protected void assertEquals(DoubleMatrix2D expected, QRDecomposition2 qr) {
    DoubleMatrix2D matComputed = LA.mult(qr.getQ(), qr.getR());
    System.out.println("Q*R: " + matComputed);
    assertEquals(expected, matComputed);
  }
  
  
  protected void assertEquals(DoubleMatrix2D expected, DoubleMatrix2D actual) {
    int iRows = expected.rows();
    int iColumns = expected.columns();
    System.out.println("Expected: " + expected);
    System.out.println("Actual: " + actual);
    
    // Safety-check
    if (iRows != actual.rows() || iColumns != actual.columns()) {
      throw new IllegalArgumentException("Both matrices must have the same dimension");
    }
    
    double dTOL = 1e-9;
    for (int iRow = 0; iRow < iRows; iRow++) {
      for (int iCol = 0; iCol < iColumns; iCol++) {
        double dExpected = expected.get(iRow, iCol);
        double dActual = actual.get(iRow, iCol);
        assertEquals(dExpected, dActual, dTOL);
      }
    }
  }

  protected void assertEquals(DoubleMatrix1D expected, DoubleMatrix1D actual) {
    int iRows = expected.size();
    int iColumns = expected.size();
    
    // Safety-check
    if (iRows != actual.size() || iColumns != actual.size()) {
      throw new IllegalArgumentException("Both vectors must have the same dimension");
    }
    
    double dTOL = 1e-9;
    for (int iRow = 0; iRow < iRows; iRow++) {
      double dExpected = expected.get(iRow);
      double dActual = actual.get(iRow);
      assertEquals(dExpected, dActual, dTOL);
    }
  }
  
  
  public void testHouseholderQR_Ex_5_4_1_Test2() {
    System.out.println();
    System.out.println("*****");
    System.out.println("Householder QR Test #2");
    System.out.println("*****");
    
    // double[][] A = new double[][] { {3, 2, 1}, {6, 5, 1}, {9, 8, 1}, {12, 11, 1}};
    double[][] A = new double[][] { {3, 2, -1}, {6, 5, +1}, {9, 8, -1}, {12, 11, 1}};
    DoubleMatrix2D matA = DoubleFactory2D2.dense.make(A);
    //DoubleMatrix2D matA = DoubleFactory2D2.dense.random(4, 3, 1000.0);
    
    System.out.println();
    System.out.println("A: " + matA);
    QRDecomposition2 qrOriginal = new QRDecomposition2(matA);
    System.out.println(qrOriginal);
    System.out.println("A (made): " + LA.mult(qrOriginal.getQ(), qrOriginal.getR()));
    
    System.out.println();
    System.out.println("A: " + matA);
    QRDecomposition2 qr = new QRDecomposition2(matA);
    System.out.println(qr);
//    System.out.println("A (made #1): " + LA.mult(qr.getQ1(), qr.getR()));
    System.out.println("A (made #2): " + LA.mult(qr.getQ2(), qr.getR()));
    System.out.println("A (made #3): " + LA.mult(qr.getQ3(), qr.getR()));
    
//    DoubleMatrix2D[] aryP = qr.aryP;
//    DoubleMatrix2D matTemp = cern.colt.matrix.DoubleFactory2D.dense.identity(4);
//    for (int i = 0; i < aryP.length; i++) {
//      System.out.println("P #" + i + ": " + aryP[i]);
//      DoubleMatrix2D matPi = cern.colt.matrix.DoubleFactory2D.dense.composeDiagonal(cern.colt.matrix.DoubleFactory2D.dense.identity(i), aryP[i]);
//      matTemp = LA.mult(matPi, matTemp);
//    }
//    System.out.println("P final: " + matTemp);
//    System.out.println("P * A: " + LA.mult(matTemp, matA));
    
    
    DoubleMatrix2D matAComputed = LA.mult(qr.getQ(), qr.getR());
    System.out.println("A (computed): " + matAComputed);
    assertEquals(matA, matAComputed);
    
    printColumnIndependence(matA, qr);
  }
  
  public void testColumnSwapping() {
    System.out.println();
    System.out.println("*****");
    System.out.println("Column Swapping Test");
    System.out.println("*****");
    
    double[][] A = new double[][] { {1, 2, 3}, {1, 5, 6}, {1, 8, 9}, {1, 11, 12}};
    DoubleMatrix2D matA = factory2D.dense.make(A);
    System.out.println("A: " + matA);
    
    int iIndex1 = 0;
    int iIndex2 = 2;
    DoubleMatrix2D matADuplicate = matA.copy();
    LA.permuteColumns(matADuplicate, iIndex1, iIndex2);
    System.out.println("A (permutted): " + matADuplicate);
    
    DoubleMatrix1D vecOriginal = matA.viewColumn(iIndex1);
    DoubleMatrix1D vecOther = matADuplicate.viewColumn(iIndex2);
    assertEquals(vecOriginal, vecOther);
    
    vecOriginal = matA.viewColumn(iIndex2);
    vecOther = matADuplicate.viewColumn(iIndex1);
    assertEquals(vecOriginal, vecOther);    
  }

  
  public void testHouseholderQR_Ex_5_4_1_Test3() {
    System.out.println();
    System.out.println("*****");
    System.out.println("Householder QR Test #3 - Column Pivoting with initial Good Matrix");
    System.out.println("*****");
    
    double[][] A = new double[][] { {3, 2, 1}, {6, 5, 1}, {9, 8, 1}, {12, 11, 1}};
    // double[][] A = new double[][] { {1, 2, 3}, {1, 5, 6}, {1, 8, 9}, {1, 11, 12}};
    DoubleMatrix2D matA = factory2D.dense.make(A);
    System.out.println("A: " + matA);
    
    QRDecomposition2 qr = new QRDecomposition2(matA, true);
    System.out.println(qr);
    
    DoubleMatrix2D matAComputed = LA.mult(qr.getQ(), qr.getR());
    System.out.println("A (original): " + matA);
    System.out.println("A (computed): " + matAComputed);
    checkDiagonalElements(qr.getR());
    //assertEquals(matA, matAComputed);
    DoubleMatrix2D matPivots = qr.getPivots();
    System.out.println("Pivots: " + matPivots);
    DoubleMatrix2D matLHS = LA.mult(matA, matPivots);
    System.out.println("A*E: " + matLHS);
    assertEquals(matLHS, matAComputed);
    
    printColumnIndependence(matA, qr);
  }

  public void testHouseholderQR_Ex_5_4_1_Test4() {
    System.out.println();
    System.out.println("*****");
    System.out.println("Householder QR Test #4 - Column Pivoting with initial bad Matrix");
    System.out.println("*****");
    
    // double[][] A = new double[][] { {3, 2, 1}, {6, 5, 1}, {9, 8, 1}, {12, 11, 1}};
    double[][] A = new double[][] { {1, 2, 3}, {1, 5, 6}, {1, 8, 9}, {1, 11, 12}};
    DoubleMatrix2D matA = factory2D.dense.make(A);
    System.out.println("A: " + matA);
    
    QRDecomposition2 qr = new QRDecomposition2(matA, true);
    System.out.println(qr);
    
    DoubleMatrix2D matAComputed = LA.mult(qr.getQ(), qr.getR());
    System.out.println("A (original): " + matA);
    System.out.println("A (computed): " + matAComputed);
    checkDiagonalElements(qr.getR());
    //assertEquals(matA, matAComputed);
    DoubleMatrix2D matPivots = qr.getPivots();
    System.out.println("Pivots: " + matPivots);
    DoubleMatrix2D matLHS = LA.mult(matA, matPivots);
    System.out.println("A*E: " + matLHS);
    assertEquals(matLHS, matAComputed);
    
    printColumnIndependence(matA, qr);
  }
  
  
  public void Atest_2x2_OK_1() {
    System.out.println();
    System.out.println("2x2 #1");
    double[][] A = new double[][] {{1,0}, {0,1}};
    DoubleMatrix2D matA = factory2D.dense.make(A);
    System.out.println("A: " + matA);
    QRDecomposition2 qr = new QRDecomposition2(matA);
    System.out.println(qr);
  }
  
  public void Atest_2x2_OK_2() {
    System.out.println();
    System.out.println("2x2 #2");
    double[][] A = new double[][] {{-1,0}, {0,1}};
    DoubleMatrix2D matA = factory2D.dense.make(A);
    System.out.println("A: " + matA);
    QRDecomposition2 qr = new QRDecomposition2(matA);
    System.out.println(qr);
  }
  
  public void Atest_2x2_Bad_3() {
    System.out.println();
    System.out.println("2x2 #3");
    double[][] A = new double[][] {{0,0}, {0,1}};
    DoubleMatrix2D matA = factory2D.dense.make(A);
    System.out.println("A: " + matA);
    QRDecomposition2 qr = new QRDecomposition2(matA);
    System.out.println(qr);
  }
  
  public void Atest_3x3_OK() {
    double[][] A = new double[][] {{0,3,0}, {2,0,0}, {0,0, 0}};
    DoubleMatrix2D matA = factory2D.dense.make(A);
    System.out.println("A: " + matA);
    QRDecomposition2 qr = new QRDecomposition2(matA);
    System.out.println(qr);
  }
  
  public void Atest_3x3_Dup() {
    double[][] A = new double[][] {{0,3,0}, {2,0,0}, {5,0,0}};
    DoubleMatrix2D matA = factory2D.dense.make(A);
    System.out.println("A: " + matA);
    QRDecomposition2 qr = new QRDecomposition2(matA);
    System.out.println(qr);
  }
  
  public void test_3x3_OK_2_No_Pivot() {
    double[][] A = new double[][] {{0,3,0}, {2,0,0}, {0,0, 0}};
    DoubleMatrix2D matA = factory2D.dense.make(A);
    System.out.println("A: " + matA);
    QRDecomposition2 qr = new QRDecomposition2(matA, false);
    System.out.println(qr);
  }
  
  public void test_3x3_OK_2_Pivot() {
    double[][] A = new double[][] {{0,3,0}, {2,0,0}, {0,0, 0}};
    DoubleMatrix2D matA = factory2D.dense.make(A);
    System.out.println("A: " + matA);
    QRDecomposition2 qr = new QRDecomposition2(matA, true);
    System.out.println(qr);
    
    checkDiagonalElements(qr.getR());
  }
  
  protected void checkDiagonalElements(DoubleMatrix2D matrix) {
    System.out.println("Checking matrix: " + matrix);
    
    int iRows = matrix.rows();
    int iCols = matrix.columns();
    int iDiags = Math.min(iRows, iCols);
    double dMaxValue = Math.abs(matrix.get(0, 0));
    for (int i = 0; i < iDiags; i++) {
      double dValue = Math.abs(matrix.get(i,i));
      assertTrue(dMaxValue >= dValue);
      dMaxValue = dValue;
    }
  }
  
  public void test_3x3_OK_2_Pivot2() {
    double[][] A = new double[][] {{0,3,0}, {2,0,0}, {0,0, 0}};
    DoubleMatrix2D matA = factory2D.dense.make(A);
    System.out.println("A: " + matA);
    QRDecomposition2 qr = new QRDecomposition2(matA, false);
    System.out.println(qr);
    
    boolean bDesiredError = false;
    try {
      checkDiagonalElements(qr.getR());
    } catch (junit.framework.AssertionFailedError errorThatWeWanted) {
      bDesiredError = true;
    }
    assertTrue("Checking to see whether our matrix was not sorted", bDesiredError);
  }
  
  public void test_3x3_OK_Random() {
    int iColumns = 3;
    DoubleMatrix2D matA = factory2D.dense.random(iColumns, iColumns);
    System.out.println("A: " + matA);
    QRDecomposition2 qr = new QRDecomposition2(matA, true);
    System.out.println(qr);
    
    checkDiagonalElements(qr.getR());
  }
  
  public void test_10x10_OK_Random() {
    int iColumns = 10;
    DoubleMatrix2D matA = DoubleFactory2D2.dense.random(iColumns, iColumns, 100);
    System.out.println("A: " + matA);
    QRDecomposition2 qr = new QRDecomposition2(matA, true);
    System.out.println(qr);
    
    checkDiagonalElements(qr.getR());
  }
  
  public void testIdentifyRedundantColumns() {
    System.out.println();
    System.out.println("*****");
    System.out.println("Householder QR Test #5 - Redundant columns");
    System.out.println("*****");
    
    double[][] A = new double[][] { {0, 1, 1}, {1, 0, 1}, {0, 0, 0}, {0, 0, 0}};
    DoubleMatrix2D matA = factory2D.dense.make(A);
    System.out.println("A: " + matA);
    
    QRDecomposition2 qr = new QRDecomposition2(matA, true);
    System.out.println(qr);
    
    DoubleMatrix2D matAComputed = LA.mult(qr.getQ(), qr.getR());
    System.out.println("A (original): " + matA);
    DoubleMatrix2D matPivots = qr.getPivots();
    System.out.println("Pivots: " + matPivots);
    DoubleMatrix2D matLHS = LA.mult(matA, matPivots);
    System.out.println("A*E: " + matLHS);
    System.out.println("A (computed): " + matAComputed);
    
    checkDiagonalElements(qr.getR());
    assertEquals(matLHS, matAComputed);
    
    printColumnIndependence(matA, qr);
  }

  
  public void testIdentifyRedundantColumns_2() {
    System.out.println();
    System.out.println("*****");
    System.out.println("Householder QR Test #6 - Redundant columns #1 - Pivoting");
    System.out.println("*****");
    
    double[][] A = new double[][] { {1, 0, 10}, {1, 11, 0}, {0, 0, 0}, {0, 0, 0}};
    DoubleMatrix2D matA = factory2D.dense.make(A);
    System.out.println("A: " + matA);
    
    QRDecomposition2 qr = new QRDecomposition2(matA, true);    
    System.out.println(qr);
    
    DoubleMatrix2D matAComputed = LA.mult(qr.getQ(), qr.getR());
    System.out.println("A (original): " + matA);
    DoubleMatrix2D matPivots = qr.getPivots();
    System.out.println("Pivots: " + matPivots);
    DoubleMatrix2D matLHS = LA.mult(matA, matPivots);
    System.out.println("A*E: " + matLHS);
    System.out.println("A (computed): " + matAComputed);
    
    checkDiagonalElements(qr.getR());
    assertEquals(matLHS, matAComputed);
    
    printColumnIndependence(matA, qr);
  }
  
  public void testIdentifyRedundantColumns_3() {
    System.out.println();
    System.out.println("*****");
    System.out.println("Householder QR Test #6 - Redundant columns #2 - No pivoting!");
    System.out.println("*****");
    
    double[][] A = new double[][] { {1, 0, 10}, {1, 11, 0}, {0, 0, 0}, {0, 0, 0}};
    DoubleMatrix2D matA = factory2D.dense.make(A);
    System.out.println("A: " + matA);
    
    QRDecomposition2 qr = new QRDecomposition2(matA, false);    
    System.out.println(qr);
    
    DoubleMatrix2D matAComputed = LA.mult(qr.getQ(), qr.getR());
    System.out.println("A (original): " + matA);
    DoubleMatrix2D matPivots = qr.getPivots();
    System.out.println("Pivots: " + matPivots);
    DoubleMatrix2D matLHS = LA.mult(matA, matPivots);
    System.out.println("A*E: " + matLHS);
    System.out.println("A (computed): " + matAComputed);
    
    assertEquals(matLHS, matAComputed);
    
    printColumnIndependence(matA, qr);
  }
  
  
  protected void printColumnIndependence(DoubleMatrix2D A, QRDecomposition2 qr) {
    // Identify redundant columns
    System.out.println();
    System.out.println("Dependent Columns: ");
    int[] iAryBadColumns = qr.getDependentColumns();
    print(iAryBadColumns);
    
    // Identify independent columns
    System.out.println();
    System.out.println("Independent Columns: ");
    int[] iAryGoodColumns = qr.getIndependentColumns();
    print(iAryGoodColumns);

    // Display the independent columns    
    DoubleMatrix2D matDependent = A.viewSelection(null, iAryGoodColumns);
    System.out.println("Independent matrix: " + matDependent);
  }
  
  public void testQR1() {
    System.out.println();
    System.out.println("*****");
    System.out.println("* Testing QRDecomposition");
    System.out.println("*****");
    
    //  0.7071   -0.7071
    // -0.8944   -0.4472
    // -0.8944   -0.4472
    
    double[][] A = new double[][] { {.7071, -.7071}, {-0.8944, -0.4472}, {-0.8944, -0.4472} };
    //double[][] A = new double[][] { {-0.8944, -0.4472}, {-0.8944, -0.4472}, {.7071, -.7071}, {-0.8944, -0.4472}};
    DoubleMatrix2D matA = factory2D.dense.make(A);    

    System.out.println("matA: " + matA);
    QRDecomposition2 qr = new QRDecomposition2(matA);
    System.out.println(qr);
    System.out.println(qr.getPivots());
    System.out.println( LA.mult(qr.getQ(), qr.getR()) );

    System.out.println();
    System.out.println();
    System.out.println("*****");
    System.out.println();
    DoubleMatrix2D matATranspose = matA.viewDice();
    //DoubleMatrix2D matATranspose = matA.viewDice().viewSelection(new int[] {0,1}, new int[] {0, 1});
    System.out.println("matA': " +  matATranspose);    
    qr = new QRDecomposition2(matATranspose, true);
    System.out.println(qr);
    System.out.println(qr.getPivots());
    System.out.println( LA.mult(qr.getQ(), qr.getR()) );
    System.out.println( LA.mult(matATranspose, qr.getPivots()) );
  }
  


  
  protected void print(int[] array) {
    for (int i = 0; i < array.length; i++) {
      System.out.println("Index #" + i + " = " + array[i]);
    }
  }
  
}